-- VGM related functions for XPMC

include globals.e
include util.e


-- Write a VGM file based on the compiled song data.
--
-- Arguments:
--
--	fname:	Filename of the VGM
--	song:	Song number (1..number of songs in mml)
--	psg:	Non-zero if PSG is used in the song
--	ym2413:	Non-zero if YM2413 is used in the song
--	ym2612:	Non-zero if YM2612 is used in the song
--
global procedure write_vgm(sequence fname, integer song, integer psg, integer ym2413, integer ym2612)
	atom factor, totalWaits, loopWaits
	integer f, fhi, machineSpeed, songSize,
	        cmd, volChange, freqChange, vol, delay, nChannels
	sequence freqTbl, oct1, fileEnding, s, vgmData, channel,
	         channelDone, channelLooped, iterations, gd3,
	         cmdPos, loopPos, lastChannelSetting

	if ym2413 then
		if length(supportedChannels) = 7 then
			if length(songs[song][5]) = 1 and
			   length(songs[song][6]) = 1 and
			   length(songs[song][7]) = 1 then
			   	-- Disable the YM2413 if none of its channels are used
				ym2413 = 0
			end if
		else
			ym2413 = 0
		end if
	end if
	
	if ym2612 then
		if length(supportedChannels) = 10 then
			if length(songs[song][5]) = 1 and
			   length(songs[song][6]) = 1 and
			   length(songs[song][7]) = 1 and
			   length(songs[song][8]) = 1 and
			   length(songs[song][9]) = 1 and
			   length(songs[song][10]) = 1 then
			   	-- Disable the YM2612 if none of its channels are used
				ym2612 = 0
			end if
		else
			ym2612 = 0
		end if
	end if
			
	nChannels = length(supportedChannels)
	
	outFile = open(fname, "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & fname, -1)
	end if

	if verbose then
		puts(1,"Generating VGM data\n")
	end if

	if updateFreq = 50 then
		machineSpeed = 3546893
	else
		machineSpeed = 3579545
	end if
	if target = TARGET_SMS and tune then
		oct1 = OCTAVE1_ALT_SMS
	else
		oct1 = OCTAVE1
	end if
	freqTbl = {0, 0, 0}

	-- Set up the frequency table for the PSG
	factor = 2.0
	freqTbl[1] = repeat(0, 12 * 6)
	for i = 2 to 7 do
		for n = 1 to 12 do
			freqTbl[1][(i - 2) * 12 + n] = floor(machineSpeed / (oct1[n] * factor * 32))
		end for
		factor *= 2.0
	end for

	-- Set up the frequency table for the YM2413
	if ym2413 then
		oct1 = OCTAVE1
		freqTbl[2] = repeat(0, 12)
		for n = 1 to 12 do
			freqTbl[2][n] = floor((oct1[n] * power(2, 18) / 50000))
		end for
	end if
	
	if ym2612 then
		freqTbl[3] = {617, 653, 692, 733, 777, 823, 872, 924, 979, 1037, 1099, 1164}
	end if
	
	vgmData = {#56, #67, #6D, #20,	-- "Vgm "
		   0, 0, 0, 0,		-- EOF offset, filled in later
		   #50, #01, 0, 0}	-- VGM version (1.50)
	vgmData &= int_to_bytes(machineSpeed)
	if ym2413 then
		vgmData &= int_to_bytes(machineSpeed)
	else
		vgmData &= int_to_bytes(0)	
	end if
	vgmData &= int_to_bytes(0)	-- GD3 tag offset
	vgmData &= int_to_bytes(0)	-- Total samples
	vgmData &= int_to_bytes(0)	-- Loop offset
	vgmData &= int_to_bytes(0)	-- Loop samples
	vgmData &= int_to_bytes(floor(updateFreq))
	vgmData &= {9, 0, 8, 0}		-- Noise feedback pattern / shift width
	if ym2612 then
		vgmData &= int_to_bytes(machineSpeed)
	else
		vgmData &= int_to_bytes(0)	
	end if
	vgmData &= int_to_bytes(0)	-- YM2151 clock (none)
	vgmData &= int_to_bytes(#0C)	-- Data offset
	vgmData &= {0, 0, 0, 0, 0, 0, 0, 0}

	totalWaits 	= 0
	loopWaits 	= 0
	channelDone 	= repeat(0, nChannels)
	channelLooped 	= repeat(0, nChannels)
	iterations 	= {0, 0}

	-- Keeps track of the numbers of waits and current size of the VGM data
	-- at the point of each command for each channel. Used for calculating
	-- the offset and length of loops in the VGM.
	cmdPos = repeat(0, nChannels)
	for i = 1 to nChannels do
		cmdPos[i] = repeat(0, length(songs[song][i]))
	end for
	loopPos = {}

	-- Used for avoiding unnecessary writes to the PSG port.
	-- Contains the last value that have been output ({frequency, volume, panning}) for
	-- each channel.
	lastChannelSetting = repeat({-1, -1, #FF}, nChannels)

	--  1 dataPtr,
	--  2 dataPos,
	--  3 delay,
	--  4 note,
	--  5 noteOffs,
	--  6 octave,
	--  7 duty,
	--  8 freq,
	--  9 freqOffs,
	-- 10 freqOffsLatch,
	-- 11 volume,
	-- 12 volOffs,
	-- 13 volOffsLatch,
	-- 14 vMac,
	-- 15 enMac,
	-- 16 en2Mac,
	-- 17 epMac,
	-- 18 mpMac,
	-- 19 loops,
	-- 20 loopIndex,
	-- 21 detune,
	-- 22 csMac,
	-- 23 mode,
	-- 24 feedback,
	-- 25 adsr,
	-- 26 operator,
	-- 27 mult,
	-- 28 ams,
	-- 29 fms,
	-- 30 lfo,
	-- 31 rateScale
	channel = repeat({song, 1, #100, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0,
	                  {0, 0, {1, 2}}, {0, 0, {1, 2}}, {0, 0, {1, 2}}, {0, 0, {1, 2}},
	                  {0, 0, 0}, {0, 0}, 0, 0, {0, #FF, {1, 2}}, 0, 0, {0, 0, 0, 0, 0}, 1, 0, 0, 0, {0, 0, 0}, 0}, nChannels)
	
	if ym2413 then
		vgmData &= {#51, #0F, #08}
		vgmData &= {#51, #02, #00}
		vgmData &= {#51, #0E, #00}
	end if
	
	if ym2612 then
		-- Turn on left and right output for all channels
		vgmData &= {#52, #B4, #C0}
		vgmData &= {#52, #B5, #C0}
		vgmData &= {#52, #B6, #C0}
		vgmData &= {#53, #B4, #C0}
		vgmData &= {#53, #B5, #C0}
		vgmData &= {#53, #B6, #C0}
		
		-- Turn off DAC, LFO
		vgmData &= {#52, #90, #00}
		vgmData &= {#52, #22, #00}
		vgmData &= {#52, #27, #00}
		vgmData &= {#52, #2B, #00}
		
		--vgmData &= {#52, #30, #71}
		--vgmData &= {#52, #34, #0D}
		--vgmData &= {#52, #38, #33}
		--vgmData &= {#52, #3C, #01}
		--vgmData &= {#53, #30, #71}
		--vgmData &= {#53, #34, #0D}
		--vgmData &= {#53, #38, #33}
		--vgmData &= {#53, #3C, #01}
	end if
	
	while sum(channelDone) != nChannels do
		for i = 1 to nChannels do
			if not channelDone[i] then
				freqChange = 0
				volChange = 0
				channel[i][3] -= #100

				-- Check if the whole part of the delay has reached 0
				if and_bits(channel[i][3], #FFFF00) = 0 then
					iterations[2] = 0
					-- Repeat until a note command has been read
					while freqChange != 2 do
						cmd = songs[channel[i][1]][i][channel[i][2]]
						cmdPos[i][channel[i][2]] = {length(vgmData), totalWaits}
						if and_bits(cmd, #F0) = CMD_NOTE or
						   and_bits(cmd, #F0) = CMD_OCTUP or
						   and_bits(cmd, #F0) = CMD_OCTDN then
							-- Do the octave change if specified
							if and_bits(cmd, #F0) = CMD_OCTUP then
								if psg and i < 5 then
									channel[i][6] += 12
								elsif ym2413 and in_range(i, 5, 7) then
									channel[i][6] += 1
								elsif ym2612 and in_range(i, 5, 10) then
									channel[i][6] += 1
								end if
							elsif and_bits(cmd, #F0) = CMD_OCTDN then
								if psg and i < 5 then
									channel[i][6] -= 12
								elsif ym2413 and in_range(i, 5, 7) then
									channel[i][6] -= 1
								elsif ym2612 and in_range(i, 5, 10) then
									channel[i][6] -= 1
								end if
							end if

							-- If the previous note was a rest we need to trigger
							-- a volume change since the channel is currently muted.
							if channel[i][4] = CMD_REST then
								volChange = 1
							end if

							-- Note number is stored in low 4 bits of the command byte.
							channel[i][4] = and_bits(cmd, #0F)
							
							channel[i][2] += 1
							
							-- Delays are 16.8 unsigned fixed point. In the song data
							-- they are stored either in two bytes (0-7FFF, for short delays)
							-- or three bytes (0-3FFFFF, for long delays).
							delay = songs[channel[i][1]][i][channel[i][2]]
							if and_bits(delay, #80) then
								channel[i][2] += 1
								delay = and_bits(delay, #7F) * #80 + songs[channel[i][1]][i][channel[i][2]]
							end if
							channel[i][2] += 1
							delay = delay * #100 + songs[channel[i][1]][i][channel[i][2]]
							
							channel[i][3] += delay

							-- The note can only be heard if the whole part of the delay
							-- is greater than zero.
							if channel[i][3] > #FF then
								freqChange = 2
							end if

							if channel[i][4] != CMD_REST then
								-- Reset volume macro
								if channel[i][14][1] then
									channel[i][14][3] = {1, 2}
									channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], channel[i][14][3])
									channel[i][11] = channel[i][14][3][3]
									volChange = 1
								end if
								-- Reset arpeggio (cumulative)
								if channel[i][15][1] then
									channel[i][15][3] = {1, 2}
									channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], channel[i][15][3])
									channel[i][5] = channel[i][15][3][3]
								end if
								-- Reset arpeggio (non-cumulative)
								if channel[i][16][1] then
									channel[i][16][3] = {1, 2}
									channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], channel[i][16][3])
									channel[i][5] = channel[i][16][3][3]
								end if
								-- Reset pitch sweep
								if channel[i][17][1] then
									channel[i][17][3] = {1, 2}
									channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], channel[i][17][3])
									channel[i][9] = channel[i][17][3][3]
								end if
								-- Reset vibrato
								if channel[i][18][1] then
									channel[i][18][3] = vibratos[2][channel[i][18][1]][2][1]
									channel[i][9] = 0
									channel[i][10] = vibratos[2][channel[i][18][1]][2][3]
								end if
								-- Reset panning
								if channel[i][22][1] and not supportsPAL then
									channel[i][22][3] = {1, 2}
									channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], channel[i][22][3])
									if channel[i][22][3][3] then
										if and_bits(channel[i][22][3][3], #80) then
											channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
											channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
										else
											channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
											channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
										end if
									else
										channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
										channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
									end if
									if psg and i < 5 and channel[i][22][2] != lastChannelSetting[i][3] then
										vgmData &= {#4F, channel[i][22][2]}
										lastChannelSetting[i][3] = channel[i][22][2]
									end if
								end if
							end if
						elsif and_bits(cmd, #F0) = CMD_OCTAVE then
							if psg and i < 5 then
								channel[i][6] = and_bits(cmd, #0F) * 12
							elsif ym2413 and in_range(i, 5, 7) then
								channel[i][6] = and_bits(cmd, #0F)
							elsif ym2612 and in_range(i, 5, 10) then
								channel[i][6] = and_bits(cmd, #0F)
							end if

						elsif and_bits(cmd, #F0) = CMD_DUTY then
							if psg and i = 4 then
								channel[i][7] = and_bits(xor_bits(cmd, 1), 1) * 4
							elsif ym2413 and in_range(i, 5, 7) then
								channel[i][7] = and_bits(cmd, #0F) * 16
							elsif ym2612 and in_range(i, 5, 10) then
								channel[i][7] = and_bits(cmd, 7)
								vgmData &= {#52 + floor((i - 5) / 3),
								            #B0 + remainder((i - 5), 3),
								            channel[i][7] + channel[i][24]}
							end if

						elsif and_bits(cmd, #F0) = CMD_VOL2 then
							channel[i][11] = and_bits(cmd, #0F)
							channel[i][14][1] = 0
							volChange = 1

						-- Callbacks are ignored when outputting to VGM
						elsif and_bits(cmd, #F0) = CMD_ARPOFF then
							channel[i][15][1] = 0
							channel[i][16][1] = 0

						elsif cmd = CMD_CBONCE or cmd = CMD_CBEVNT then
							channel[i][2] += 1

						elsif cmd = CMD_MULT then
							channel[i][2] += 1
							channel[i][27] = songs[channel[i][1]][i][channel[i][2]]
							if ym2612 and in_range(i, 5, 10) then
								if channel[i][26] then
									vgmData &= {#52 + floor((i - 5) / 3),
										    #30 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][21] * #10 + channel[i][27]}
								else
									for j = 0 to 3 do
										vgmData &= {#52 + floor((i - 5) / 3),
											    #30 + j*4 + remainder((i - 5), 3),
											    channel[i][21] * #10 + channel[i][27]}
									end for
								end if
							end if
							
						elsif and_bits(cmd, #F0) = CMD_MODE then
							channel[i][23] = and_bits(cmd, #0F)

						elsif and_bits(cmd, #F0) = CMD_OPER then
							channel[i][26] = and_bits(cmd, #0F)
							
						elsif and_bits(cmd, #F0) = CMD_FEEDBK then
							if ym2612 and in_range(i, 5, 10) then
								channel[i][24] = and_bits(cmd, 7) * 8
								vgmData &= {#52 + floor((i - 5) / 3),
								            #B0 + remainder((i - 5), 3),
								            channel[i][7] + channel[i][24]}
							end if
							--? vgmData[length(vgmData)-2..length(vgmData)]
							
						elsif cmd = CMD_ADSR then
							channel[i][2] += 1
							if ym2413 and in_range(i, 5, 7) then
								channel[i][25] = adsrs[2][songs[channel[i][1]][i][channel[i][2]]][2]
								vgmData &= {#51, #04, channel[i][25][1] * #10 + channel[i][25][2]}
								vgmData &= {#51, #05, channel[i][25][1] * #10 + channel[i][25][2]}
								vgmData &= {#51, #06, channel[i][25][3] * #10 + channel[i][25][4]}
								vgmData &= {#51, #07, channel[i][25][3] * #10 + channel[i][25][4]}
							elsif ym2612 and in_range(i, 5, 10) then
								f = and_bits(channel[i][25][2], #80)
								channel[i][25] = adsrs[2][songs[channel[i][1]][i][channel[i][2]]][2]
								channel[i][25][2] = or_bits(channel[i][25][2], f)
								if channel[i][26] then
									vgmData &= {#52 + floor((i - 5) / 3),
										    #50 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][25][1] + channel[i][31]}
									vgmData &= {#52 + floor((i - 5) / 3),
										    #60 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][25][2]}
									vgmData &= {#52 + floor((i - 5) / 3),
										    #70 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][25][3]}
									vgmData &= {#52 + floor((i - 5) / 3),
										    #80 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][25][4]}
								else
									for j = 0 to 3 do
										vgmData &= {#52 + floor((i - 5) / 3),
											    #50 + j*4 + remainder((i - 5), 3),
											    channel[i][25][1] + channel[i][31]}
										vgmData &= {#52 + floor((i - 5) / 3),
											    #60 + j*4 + remainder((i - 5), 3),
											    channel[i][25][2]}
										vgmData &= {#52 + floor((i - 5) / 3),
											    #70 + j*4 + remainder((i - 5), 3),
											    channel[i][25][3]}
										vgmData &= {#52 + floor((i - 5) / 3),
											    #80 + j*4 + remainder((i - 5), 3),
											    channel[i][25][4]}
									end for
								end if
								--? vgmData[length(vgmData)-11..length(vgmData)]
							end if
						
						elsif cmd = CMD_RSCALE then
							channel[i][2] += 1
							channel[i][31] = songs[channel[i][1]][i][channel[i][2]] * #40
							if channel[i][26] then
								vgmData &= {#52 + floor((i - 5) / 3),
									    #50 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
									    channel[i][25][1] + channel[i][31]}
							else
								for j = 0 to 3 do
									vgmData &= {#52 + floor((i - 5) / 3),
										    #50 + j*4 + remainder((i - 5), 3),
										    channel[i][25][1] + channel[i][31]}
								end for
							end if
							
						elsif cmd = CMD_VOLSET then
							channel[i][2] += 1
							channel[i][11] = songs[channel[i][1]][i][channel[i][2]]
							channel[i][14][1] = 0
							volChange = 1
							
						elsif cmd = CMD_VOLMAC then
							channel[i][2] += 1
							channel[i][14][1] = songs[channel[i][1]][i][channel[i][2]]
							channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], {1, 2})
							channel[i][11] = channel[i][14][3][3]
							volChange = 1

						elsif cmd = CMD_PANMAC then
							channel[i][2] += 1
							channel[i][22][1] = songs[channel[i][1]][i][channel[i][2]]
							channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], {1, 2})
							if channel[i][22][3][3] then
								if and_bits(channel[i][22][3][3], #80) then
									channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
									channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
								else
									channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
									channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
								end if
							else
								channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
								channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
							end if
							if psg and i < 5 and channel[i][22][2] != lastChannelSetting[i][3] and not supportsPAL then
								vgmData &= {#4F, channel[i][22][2]}
								lastChannelSetting[i][3] = channel[i][22][2]
							end if

						elsif cmd = CMD_ARPMAC then
							channel[i][2] += 1
							channel[i][15][1] = songs[channel[i][1]][i][channel[i][2]]
							channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], {1, 2})
							channel[i][5] = channel[i][15][3][3]
							channel[i][16][1] = 0
							freqChange = 1

						elsif cmd = CMD_DUTMAC then
							channel[i][2] += 1

						elsif cmd = CMD_DETUNE then
							channel[i][2] += 1
							channel[i][21] = songs[channel[i][1]][i][channel[i][2]]
							if ym2612 and in_range(i, 5, 10) then
								if channel[i][26] then
									vgmData &= {#52 + floor((i - 5) / 3),
										    #30 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][21] * #10 + channel[i][27]}
								else
									for j = 0 to 3 do
										vgmData &= {#52 + floor((i - 5) / 3),
											    #30 + j*4 + remainder((i - 5), 3),
											    channel[i][21] * #10 + channel[i][27]}
									end for
								end if
							end if

						elsif cmd = CMD_MODMAC then
							channel[i][2] += 1
							channel[i][30][1] = songs[channel[i][1]][i][channel[i][2]]
							if channel[i][30][1] then
								channel[i][30][3] = mods[2][channel[i][30][1]][2]
								vgmData &= {#52, #22, channel[i][30][3][1] + 8}
								vgmData &= {#52 + floor((i - 5) / 3),
								            #B4 + remainder((i - 5), 3),
								            #C0 + (channel[i][30][3][2] * 8) + channel[i][30][3][3]}
							else
								vgmData &= {#52, #22, 0}
							end if
						
						elsif cmd = CMD_HWAM then
							channel[i][2] += 1
							channel[i][25][2] = and_bits(channel[i][25][2], #7F)
							channel[i][25][2] = or_bits(channel[i][25][2], songs[channel[i][1]][i][channel[i][2]] * #80)
							if ym2612 and in_range(i, 5, 10) then
								if channel[i][26] then
									vgmData &= {#52 + floor((i - 5) / 3),
										    #60 + (channel[i][26] - 1)*4 + remainder((i - 5), 3),
										    channel[i][25][2]}
								else
									for j = 0 to 3 do
										vgmData &= {#52 + floor((i - 5) / 3),
											    #60 + j*4 + remainder((i - 5), 3),
											    channel[i][25][2]}
									end for
								end if
							end if
							
						elsif cmd = CMD_APMAC2 then
							channel[i][2] += 1
							channel[i][16][1] = songs[channel[i][1]][i][channel[i][2]]
							channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], {1, 2})
							channel[i][15][1] = 0
							freqChange = 1

						elsif cmd = CMD_SWPMAC then
							channel[i][2] += 1
							channel[i][17][1] = songs[channel[i][1]][i][channel[i][2]]
							if channel[i][17][1] then
								channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], {1, 2})
								channel[i][9] = 0
								freqChange = 1
							else
								channel[i][9] = 0
								freqChange = 1
							end if	

						elsif cmd = CMD_VIBMAC then
							channel[i][2] += 1
							channel[i][18][1] = songs[channel[i][1]][i][channel[i][2]]
							if channel[i][18][1] then
								channel[i][18][3] = vibratos[2][channel[i][18][1]][2][1]
								channel[i][9] = 0
								channel[i][10] = vibratos[2][channel[i][18][1]][2][3]
								freqChange = 1
							else
								freqChange = 1
								channel[i][9] = 0
							end if

						elsif cmd = CMD_JMP then
							channel[i][2] = songs[channel[i][1]][i][channel[i][2] + 1] +
									songs[channel[i][1]][i][channel[i][2] + 2] * #100
							channelLooped[i] = 1
							if sum(channelLooped) = nChannels then
								loopPos = cmdPos[i][channel[i][2] + 1]
								channelDone = repeat(1, nChannels)
								freqChange = 2
							end if

						elsif cmd = CMD_DJNZ then
							channel[i][19][channel[i][20]] -= 1
							if channel[i][19][channel[i][20]] != 0 then
								channel[i][2] = songs[channel[i][1]][i][channel[i][2] + 1] +
										songs[channel[i][1]][i][channel[i][2] + 2] * #100
							else
								channel[i][20] -= 1
								channel[i][2] += 2
							end if

						elsif cmd = CMD_LOPCNT then
							channel[i][2] += 1
							channel[i][20] += 1
							channel[i][19][channel[i][20]] = songs[channel[i][1]][i][channel[i][2]]

						elsif cmd = CMD_END then
							channel[i][4] = cmd
							channelDone[i] = 1
							freqChange = 2

						end if

						channel[i][2] += 1
						iterations[2] += 1
						if iterations[2] = 131072 then
							puts(1, "Internal error: Appears to be stuck in an infinite loop. Exiting\n")
							? {i, channel[1][2], channel[2][2], channel[3][2], channel[4][2]}
							exit
						end if
					end while
				else

					-- Update effects as needed
					if channel[i][14][1] then
						channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], channel[i][14][3])
						channel[i][11] = channel[i][14][3][3]
						volChange = 1
					end if
					if channel[i][15][1] then
						channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], channel[i][15][3])
						channel[i][5] += channel[i][15][3][3]
						freqChange = 1
					end if
					if channel[i][16][1] then
						channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], channel[i][16][3])
						channel[i][5] = channel[i][16][3][3]
						freqChange = 1
					end if
					if channel[i][17][1] then
						channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], channel[i][17][3])
						channel[i][9] += channel[i][17][3][3]
						freqChange = 1
					end if
					if channel[i][18][1] then
						if channel[i][18][3] = 0 then
							channel[i][9] = channel[i][10]
							channel[i][10] = -channel[i][10]
							channel[i][18][3] = vibratos[2][channel[i][18][1]][2][2]
							freqChange = 1
						end if
						channel[i][18][3] -= 1
					end if
					if channel[i][22][1] and not supportsPAL then
						channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], channel[i][22][3])
						if channel[i][22][3][3] then
							if and_bits(channel[i][22][3][3], #80) then
								channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
								channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
							else
								channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
								channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
							end if
						else
							channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
							channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
						end if
						if psg and i < 5 and channel[i][22][2] != lastChannelSetting[i][3] then
							vgmData &= {#4F, channel[i][22][2]}
							lastChannelSetting[i][3] = channel[i][22][2]
						end if
					end if
				end if
				if channel[i][4] = CMD_END or channel[i][4] = CMD_REST then
					if lastChannelSetting[i][2] != 0 then
						if psg and i < 5 then
							vgmData &= {#50, or_bits(#9F, (i - 1) * #20)}
							vgmData &= {#50, or_bits(#1F, (i - 1) * #20)}
							lastChannelSetting[i] = {-1, 0, lastChannelSetting[i][3]}
						elsif ym2413 and in_range(i, 5, 7) then
							if channel[i][23] = 0 then
								--vgmData &= {#51, #30 + (i - 5), channel[i][7] + #0F}
								vgmData &= {#51, #20 + (i - 5), floor(channel[i][8] / #100)}
								--lastChannelSetting[i] = {-1, 0, lastChannelSetting[i][3]}
							end if
						elsif ym2612 and in_range(i, 5, 10) then
							if channel[i][23] = 0 then
								vgmData &= {#52, -- + floor((i - 5) / 3),
								            #28,
								            i +floor((i - 5) / 3)- 5} --remainder(i - 5, 3)}
								--lastChannelSetting[i] = {-1, lastChannelSetting[i][2], lastChannelSetting[i][3]}
							end if
						end if
					end if
					if channel[i][4] = CMD_END then
						channelDone[i] = 1
						channelLooped[i] = 1
					end if
				else	
					if freqChange then
						if i < 4 and psg then
							if channel[i][6] + channel[i][4] + channel[i][5] < 0 then
								channel[i][8] = freqTbl[1][1]
							elsif channel[i][6] + channel[i][4] + channel[i][5] > 59 then
								channel[i][8] = freqTbl[1][60]
							else
								channel[i][8] = freqTbl[1][channel[i][6] + channel[i][4] + channel[i][5] + 1]
							end if
							channel[i][8] -= channel[i][9] + channel[i][21]

							if channel[i][8] >= #03EF then
								channel[i][8] = #03EF
							elsif channel[i][8] < #001C then
								channel[i][8] = #001C
							end if
							if channel[i][8] != lastChannelSetting[i][1] then
								vgmData &= {#50, or_bits(#80 + (i - 1) * #20, and_bits(channel[i][8], #0F))}
								vgmData &= {#50, floor(and_bits(channel[i][8], #3F0) / #10)}
								lastChannelSetting[i][1] = channel[i][8]
							end if
						elsif i = 4 and psg then
							vgmData &= {#50, #E0 + channel[i][7] + and_bits(channel[i][4], 3)}
							vgmData &= {#50, #60 + channel[i][7] + and_bits(channel[i][4], 3)}
						
						elsif in_range(i, 5, 7) and ym2413 then
							if freqChange = 2 then
								-- KEY_OFF
								vgmData &= {#51, #20 + (i - 5), floor(channel[i][8] / #100)}
							end if
							f = (channel[i][6] + 1) * 12 + channel[i][4] + channel[i][5]
							if f < 0 then
								channel[i][8] = freqTbl[2][1]
								fhi = 0
							elsif f > 95 then
								channel[i][8] = freqTbl[2][12]
								fhi = 7 * #200
							else
								channel[i][8] = freqTbl[2][remainder(f, 12) + 1]
								fhi = floor(f / 12) * #200
							end if
							channel[i][8] += channel[i][9] + channel[i][21]

							if channel[i][8] >= 511 then
								channel[i][8] = 511
							elsif channel[i][8] < 0 then
								channel[i][8] = 0
							end if
							channel[i][8] += fhi
							
							if channel[i][8] != lastChannelSetting[i][1] then
								vgmData &= {#51, #10 + (i - 5), and_bits(channel[i][8], #FF)}
								lastChannelSetting[i][1] = channel[i][8]
							end if
							vgmData &= {#51, #20 + (i - 5), floor(channel[i][8] / #100) + #30}
							--printf(1, "Ch%d: Writing %02x to %02x, ", {i-5, and_bits(channel[i][8], #FF), #10 + (i - 5)})
							--printf(1, "%02x to %02x\n", {floor(channel[i][8] / #100) + #30, #20 + (i - 5)})
						
						elsif in_range(i, 5, 10) and ym2612 then
							if freqChange = 2 then
								-- KEY_OFF
								vgmData &= {#52, 
								            #28,
								            i + floor((i - 5) / 3) - 5} 
							end if
							f = (channel[i][6] + 1) * 12 + channel[i][4] + channel[i][5]
							if f < 0 then
								channel[i][8] = freqTbl[2][1]
							elsif f > 95 then
								channel[i][8] = freqTbl[2][12] + 7 * #800
							else
								channel[i][8] = freqTbl[3][remainder(f, 12) + 1] + floor(f / 12) * #800
							end if
							if channel[i][8] != lastChannelSetting[i][1] then
								vgmData &= {#52 + floor((i - 5) / 3),
								            #A4 + remainder((i - 5), 3),
								            floor(channel[i][8] / #100)}
								vgmData &= {#52 + floor((i - 5) / 3),
								            #A0 + remainder((i - 5), 3),
								            and_bits(channel[i][8], #FF)}
								--? vgmData[length(vgmData)-5..length(vgmData)]
								lastChannelSetting[i][1] = channel[i][8]
							end if
							vgmData &= {#52, 
							            #28,
							            #F0 + i +floor((i - 5) / 3)- 5} 
							--? vgmData[length(vgmData)-2..length(vgmData)]
						end if
					end if
					
					if volChange then
						vol = channel[i][11] + channel[i][12]
						if vol != lastChannelSetting[i][2] then
							if psg and i < 5 then
								if vol > 15 then
									vol = 15
								elsif vol < 0 then
									vol = 0
								end if
								vgmData &= {#50, or_bits(#90 + xor_bits(vol, 15), (i - 1) * #20)}
								vgmData &= {#50, or_bits(#10 + xor_bits(vol, 15), (i - 1) * #20)}
							elsif ym2413 and in_range(i, 5, 7) then
								if vol > 15 then
									vol = 15
								elsif vol < 0 then
									vol = 0
								end if
								if channel[i][23] = 0 then
									vgmData &= {#51, #30 + (i - 5), channel[i][7] + xor_bits(vol, 15)}
									printf(1, "Ch%d: Writing %02x to %02x\n", {i-5, channel[i][7] + xor_bits(vol, 15), #30 + (i - 5)})
								end if
							elsif ym2612 and in_range(i, 5, 10) then
								if vol > 127 then
									vol = 127
								elsif vol < 0 then
									vol = 0
								end if
								if channel[i][23] = 0 then
									vgmData &= {#52 + floor((i - 5) / 3),
									            #40 + remainder((i - 5), 3),
									            xor_bits(vol, 127)}
									vgmData &= {#52 + floor((i - 5) / 3),
									            #44 + remainder((i - 5), 3),
									            xor_bits(vol, 127)}
									vgmData &= {#52 + floor((i - 5) / 3),
									            #48 + remainder((i - 5), 3),
									            xor_bits(vol, 127)}
									vgmData &= {#52 + floor((i - 5) / 3),
									            #4C + remainder((i - 5), 3),
									            xor_bits(vol, 127)}
									--? vgmData[length(vgmData)-11..length(vgmData)]
								end if
							end if
							lastChannelSetting[i][2] = vol
						end if
					end if
				end if
			end if
		end for
		-- Wait one frame
		if updateFreq = 50 then
			vgmData &= #63
			totalWaits += 882
		else
			vgmData &= #62
			totalWaits += 735
		end if
		iterations[1] += 1
		if iterations[1] = 54000 then
			puts(1, "Internal error: Appears to be stuck in an infinite loop. Exiting\n")
			? {channel[1][2], channel[2][2], channel[3][2], channel[4][2]}
			exit
		end if
	end while
	--? vgmData[#40..length(vgmData)]
	
	if length(loopPos) then
		vgmData = vgmData[1..#18] &
			  int_to_bytes(totalWaits) &
			  int_to_bytes(loopPos[1] - #1C) &
			  int_to_bytes(totalWaits - loopPos[2]) &
			  vgmData[#25..length(vgmData)]
		loopPos[2] = totalWaits - loopPos[2]
	else
		vgmData = vgmData[1..#18] & int_to_bytes(totalWaits) & vgmData[#1D..length(vgmData)]
		loopPos = {0, 0}
	end if	
	vgmData &= #66

	-- Set EOF offset and GD3 offset
	vgmData = vgmData[1..4] & int_to_bytes(length(vgmData) - 4) & vgmData[9..length(vgmData)]
	vgmData = vgmData[1..#14] & int_to_bytes(length(vgmData) - #14) & vgmData[#19..length(vgmData)]

	puts(outFile, vgmData)

	-- Create GD3 tag
	s = date()
	gd3 = ascii_to_wide(songTitle & 0 & 0)
	gd3 &= ascii_to_wide("Unknown" & 0 & 0)
	if supportsPAL then
		gd3 &= ascii_to_wide("Sega Master System" & 0 & 0)
	else
		gd3 &= ascii_to_wide("Sega Game Gear" & 0 & 0)
	end if
	gd3 &= ascii_to_wide(songProgrammer & 0 & 0)
	gd3 &= ascii_to_wide(sprintf("%04d/%02d/%02d", {s[1] + 1900, s[2], s[3]}) & 0)
	gd3 &= ascii_to_wide("XPMC" & 0)
	gd3 &= ascii_to_wide("Composer: " & songComposer & 0)
	gd3 = "Gd3 " &
	      {0, 1, 0, 0} &
	      int_to_bytes(length(gd3)) &
	      gd3
	puts(outFile, gd3)

	close(outFile)

	if verbose then
		for i = 1 to nChannels do
			printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d / %d ticks\n", {1, songLen[1][i], songLoopLen[1][i]})
		end for
		printf(1,"VGM size: %d bytes + %d bytes GD3\nVGM length: %d / %d seconds\n", {length(vgmData), length(gd3), floor(totalWaits / 44100), floor(loopPos[2] / 44100)})
	end if
end procedure
